from __future__ import with_statement
import unittest

import random, time, threading

from whoosh import analysis, fields, query, writing
from whoosh.filedb.filestore import RamStorage
from whoosh.filedb.filetables import TermIndexReader
from whoosh.support.testing import TempIndex, TempStorage


class TestWriting(unittest.TestCase):
    def test_no_stored(self):
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        with TempIndex(schema, "nostored") as ix:
            domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                      u"golf", u"hotel", u"india")
            
            w = ix.writer()
            for i in xrange(20):
                w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
            w.commit()
            
            with ix.reader() as r:
                self.assertEqual(sorted([int(id) for id in r.lexicon("id")]), range(20))
    
    def test_asyncwriter(self):
        schema = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        with TempIndex(schema, "asyncwriter") as ix:
            domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                      u"golf", u"hotel", u"india")
            
            writers = []
            # Simulate doing 20 (near-)simultaneous commits. If we weren't using
            # AsyncWriter, at least some of these would fail because the first
            # writer wouldn't be finished yet.
            for i in xrange(20):
                w = writing.AsyncWriter(ix)
                writers.append(w)
                w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
                w.commit()
            
            # Wait for all writers to finish before checking the results
            for w in writers:
                if w.running:
                    w.join()
            
            # Check whether all documents made it into the index.
            with ix.reader() as r:
                self.assertEqual(sorted([int(id) for id in r.lexicon("id")]), range(20))
        
    def test_asyncwriter_no_stored(self):
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        with TempIndex(schema, "asyncnostored") as ix:
            domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                      u"golf", u"hotel", u"india")
            
            writers = []
            # Simulate doing 20 (near-)simultaneous commits. If we weren't using
            # AsyncWriter, at least some of these would fail because the first
            # writer wouldn't be finished yet.
            for i in xrange(20):
                w = writing.AsyncWriter(ix)
                writers.append(w)
                w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
                w.commit()
            
            # Wait for all writers to finish before checking the results
            for w in writers:
                if w.running:
                    w.join()
            
            # Check whether all documents made it into the index.
            with ix.reader() as r:
                self.assertEqual(sorted([int(id) for id in r.lexicon("id")]), range(20))
        
    def test_buffered(self):
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        with TempIndex(schema, "buffered") as ix:
            domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                      u"golf", u"hotel", u"india")
            
            w = writing.BufferedWriter(ix, period=None, limit=10,
                                       commitargs={"merge": False})
            for i in xrange(100):
                w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
            time.sleep(0.5)
            w.close()
            
            self.assertEqual(len(ix._segments()), 10)
    
    def test_buffered_search(self):
        schema = fields.Schema(id=fields.STORED, text=fields.TEXT)
        with TempIndex(schema, "bufferedsearch") as ix:
            w = writing.BufferedWriter(ix, period=None, limit=5)
            w.add_document(id=1, text=u"alfa bravo charlie")
            w.add_document(id=2, text=u"bravo tango delta")
            w.add_document(id=3, text=u"tango delta echo")
            w.add_document(id=4, text=u"charlie delta echo")
            
            with w.searcher() as s:
                r = s.search(query.Term("text", u"tango"), scored=False)
                self.assertEqual([d["id"] for d in r], [2, 3])
                
            w.add_document(id=5, text=u"foxtrot golf hotel")
            w.add_document(id=6, text=u"india tango juliet")
            w.add_document(id=7, text=u"tango kilo lima")
            w.add_document(id=8, text=u"mike november echo")
            
            with w.searcher() as s:
                r = s.search(query.Term("text", u"tango"), scored=False)
                self.assertEqual([d["id"] for d in r], [2, 3, 6, 7])
                
            w.close()
    
    def test_buffered_update(self):
        schema = fields.Schema(id=fields.ID(stored=True, unique=True),
                               payload=fields.STORED)
        with TempIndex(schema, "bufferedupdate") as ix:
            w = writing.BufferedWriter(ix, period=None, limit=5)
            for i in xrange(10):
                for char in u"abc":
                    fs = dict(id=char, payload=unicode(i) + char)
                    w.update_document(**fs)
                    
            with w.reader() as r:
                self.assertEqual(sorted(r.all_stored_fields(), key=lambda x: x["id"]),
                                 [{'id': u'a', 'payload': u'9a'},
                                  {'id': u'b', 'payload': u'9b'},
                                  {'id': u'c', 'payload': u'9c'}])
                self.assertEqual(r.doc_count(), 3)
                
            w.close()
    
    def test_buffered_threads(self):
        class SimWriter(threading.Thread):
            def __init__(self, w, domain):
                threading.Thread.__init__(self)
                self.w = w
                self.domain = domain
                
            def run(self):
                w = self.w
                domain = self.domain
                for _ in xrange(10):
                    w.update_document(name=random.choice(domain))
                    time.sleep(random.uniform(0.01, 0.1))
        
        schema = fields.Schema(name=fields.ID(unique=True, stored=True))
        with TempIndex(schema, "buffthreads") as ix:
            domain = u"alfa bravo charlie delta".split()
            w = writing.BufferedWriter(ix, limit=10)
            threads = [SimWriter(w, domain) for _ in xrange(10)]
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
            w.close()
            
            with ix.reader() as r:
                self.assertEqual(r.doc_count(), 4)
                self.assertEqual(sorted([d["name"] for d in r.all_stored_fields()]),
                                 domain)
    
    def test_fractional_weights(self):
        ana = analysis.RegexTokenizer(r"\S+") | analysis.DelimitedAttributeFilter()
        
        # With Positions format
        schema = fields.Schema(f=fields.TEXT(analyzer=ana))
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        w.add_document(f=u"alfa^0.5 bravo^1.5 charlie^2.0 delta^1.5")
        w.commit()
        
        with ix.searcher() as s:
            wts = []
            for word in s.lexicon("f"):
                p = s.postings("f", word)
                wts.append(p.weight())
            self.assertEqual(wts, [0.5, 1.5, 2.0, 1.5])
        
        # Try again with Frequency format
        schema = fields.Schema(f=fields.TEXT(analyzer=ana, phrase=False))
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        w.add_document(f=u"alfa^0.5 bravo^1.5 charlie^2.0 delta^1.5")
        w.commit()
        
        with ix.searcher() as s:
            wts = []
            for word in s.lexicon("f"):
                p = s.postings("f", word)
                wts.append(p.weight())
            self.assertEqual(wts, [0.5, 1.5, 2.0, 1.5])
            
    def test_cancel_delete(self):
        schema = fields.Schema(id=fields.ID(stored=True))
        # Single segment
        with TempIndex(schema, "canceldelete1") as ix:
            w = ix.writer()
            for char in u"ABCD":
                w.add_document(id=char)
            w.commit()
            
            with ix.reader() as r:
                self.assertFalse(r.has_deletions())
            
            w = ix.writer()
            w.delete_document(2)
            w.delete_document(3)
            w.cancel()
            
            with ix.reader() as r:
                self.assertFalse(r.has_deletions())
                self.assertFalse(r.is_deleted(2))
                self.assertFalse(r.is_deleted(3))
        
        # Multiple segments
        with TempIndex(schema, "canceldelete2") as ix:
            for char in u"ABCD":
                w = ix.writer()
                w.add_document(id=char)
                w.commit(merge=False)
            
            with ix.reader() as r:
                self.assertFalse(r.has_deletions())
            
            w = ix.writer()
            w.delete_document(2)
            w.delete_document(3)
            w.cancel()
            
            with ix.reader() as r:
                self.assertFalse(r.has_deletions())
                self.assertFalse(r.is_deleted(2))
                self.assertFalse(r.is_deleted(3))
    
    def test_delete_nonexistant(self):
        from whoosh.writing import IndexingError
        
        schema = fields.Schema(id=fields.ID(stored=True))
        # Single segment
        with TempIndex(schema, "deletenon1") as ix:
            w = ix.writer()
            for char in u"ABC":
                w.add_document(id=char)
            w.commit()
            
            try:
                w = ix.writer()
                self.assertRaises(IndexingError, w.delete_document, 5)
            finally:
                w.cancel()
        
        # Multiple segments
        with TempIndex(schema, "deletenon1") as ix:
            for char in u"ABC":
                w = ix.writer()
                w.add_document(id=char)
                w.commit(merge=False)
            
            try:
                w = ix.writer()
                self.assertRaises(IndexingError, w.delete_document, 5)
            finally:
                w.cancel()
            
    def test_read_inline(self):
        schema = fields.Schema(a=fields.TEXT)
        self.assertTrue(schema["a"].scorable)
        with TempIndex(schema, "readinline") as ix:
            w = ix.writer()
            w.add_document(a=u"alfa")
            w.add_document(a=u"bravo")
            w.add_document(a=u"charlie")
            w.commit()
            
            tr = TermIndexReader(ix.storage.open_file("_readinline_1.trm"))
            for i, item in enumerate(tr.items()):
                self.assertEqual(item[1][1], ((i,), (1.0,),
                                              ('\x00\x00\x00\x01]q\x01K\x00a',),
                                              1.0, 1))
            tr.close()
            
            with ix.reader() as r:
                pr = r.postings("a", "bravo")
                self.assertEqual(pr.id(), 1)
            
        

if __name__ == '__main__':
    unittest.main()



"""
Contains functions implementing the Levenshtein distance algorithm.
"""


def relative(a, b):
    """Returns the relative distance between two strings, in the range
    [0-1] where 1 means total equality.
    """
    
    d = distance(a, b)
    longer = float(max((len(a), len(b))))
    shorter = float(min((len(a), len(b))))    
    r = ((longer - d) / longer) * (shorter / longer)
    return r


def distance(s, t):
    """Returns the Levenshtein edit distance between two strings."""
    
    m, n = len(s), len(t)
    d = [range(n + 1)]
    d += [[i] for i in range(1, m + 1)]
    for i in range(0, m):
        for j in range(0, n):
            cost = 1
            if s[i] == t[j]:
                cost = 0
            d[i + 1].append(min(d[i][j + 1] + 1,    # deletion
                                d[i + 1][j] + 1,    # insertion
                                d[i][j] + cost))    # substitution
    return d[m][n]




# encoding: utf8
from __future__ import with_statement
import unittest
import random
from os import mkdir
from os.path import exists
from shutil import rmtree

from whoosh.filedb.filestore import RamStorage
from whoosh.filedb.filetables import (HashReader, HashWriter,
                                      OrderedHashWriter, OrderedHashReader,
                                      StoredFieldWriter, StoredFieldReader,
                                      TermIndexWriter, TermIndexReader)
from whoosh.support.testing import TempStorage


class TestTables(unittest.TestCase):
    def randstring(self, domain, minlen, maxlen):
        return "".join(random.sample(domain, random.randint(minlen, maxlen)))
    
    def test_termkey(self):
        with TempStorage("termkey") as st:
            tw = TermIndexWriter(st.create_file("test.trm"))
            tw.add(("alfa", u"bravo"), (1.0, 2, 3))
            tw.add((u"alfa", u"Ã¦Ã¯Å�Ãº"), (4.0, 5, 6))
            tw.add((u"text", u"æ—¥æœ¬èªž"), (7.0, 8, 9))
            tw.close()
            
            tr = TermIndexReader(st.open_file("test.trm"))
            self.assertTrue(("alfa", u"bravo") in tr)
            self.assertTrue((u"alfa", u"Ã¦Ã¯Å�Ãº") in tr)
            self.assertTrue((u"text", u"æ—¥æœ¬èªž") in tr)
            tr.close()
        
    def test_random_termkeys(self):
        def random_fieldname():
            return "".join(chr(random.randint(65, 90)) for _ in xrange(1, 20))
        
        def random_token():
            return "".join(unichr(random.randint(0, 0xd7ff)) for _ in xrange(1, 20))
        
        domain = sorted([(random_fieldname(), random_token()) for _ in xrange(1000)])
        
        st = RamStorage()
        tw = TermIndexWriter(st.create_file("test.trm"))
        for term in domain:
            tw.add(term, (1.0, 0, 1))
        tw.close()
        
        tr = TermIndexReader(st.open_file("test.trm"))
        for term in domain:
            self.assertTrue(term in tr)
        
    def test_hash(self):
        with TempStorage("hash") as st:
            hwf = st.create_file("test.hsh")
            hw = HashWriter(hwf)
            hw.add("foo", "bar")
            hw.add("glonk", "baz")
            hw.close()
            
            hrf = st.open_file("test.hsh")
            hr = HashReader(hrf)
            self.assertEqual(hr.get("foo"), "bar")
            self.assertEqual(hr.get("baz"), None)
            hr.close()
    
    def test_hash_contents(self):
        samp = set((('alfa', 'bravo'), ('charlie', 'delta'), ('echo', 'foxtrot'),
                   ('golf', 'hotel'), ('india', 'juliet'), ('kilo', 'lima'),
                   ('mike', 'november'), ('oskar', 'papa'), ('quebec', 'romeo'),
                   ('sierra', 'tango'), ('ultra', 'victor'), ('whiskey', 'xray')))
        
        with TempStorage("hashcontents") as st:
            hwf = st.create_file("test.hsh")
            hw = HashWriter(hwf)
            hw.add_all(samp)
            hw.close()
            
            hrf = st.open_file("test.hsh")
            hr = HashReader(hrf)
            self.assertEqual(set(hr.items()), samp)
            hr.close()
        
    def test_random_hash(self):
        with TempStorage("randomhash") as st:
            domain = "abcdefghijklmnopqrstuvwxyz"
            domain += domain.upper()
            times = 1000
            minlen = 1
            maxlen = len(domain)
            
            samp = dict((self.randstring(domain, minlen, maxlen),
                         self.randstring(domain, minlen, maxlen)) for _ in xrange(times))
            
            hwf = st.create_file("test.hsh")
            hw = HashWriter(hwf)
            for k, v in samp.iteritems():
                hw.add(k, v)
            hw.close()
            
            keys = samp.keys()
            random.shuffle(keys)
            hrf = st.open_file("test.hsh")
            hr = HashReader(hrf)
            for k in keys:
                v = hr[k]
                self.assertEqual(v, samp[k])
            hr.close()
    
    def test_ordered_hash(self):
        times = 10000
        with TempStorage("orderedhash") as st:
            hwf = st.create_file("test.hsh")
            hw = HashWriter(hwf)
            hw.add_all(("%08x" % x, str(x)) for x in xrange(times))
            hw.close()
            
            keys = range(times)
            random.shuffle(keys)
            hrf = st.open_file("test.hsh")
            hr = HashReader(hrf)
            for x in keys:
                self.assertEqual(hr["%08x" % x], str(x))
            hr.close()
        
    def test_ordered_closest(self):
        keys = ['alfa', 'bravo', 'charlie', 'delta', 'echo', 'foxtrot', 'golf',
                'hotel', 'india', 'juliet', 'kilo', 'lima', 'mike', 'november']
        values = [''] * len(keys)
        
        with TempStorage("orderedclosest") as st:
            hwf = st.create_file("test.hsh")
            hw = OrderedHashWriter(hwf)
            hw.add_all(zip(keys, values))
            hw.close()
            
            hrf = st.open_file("test.hsh")
            hr = OrderedHashReader(hrf)
            ck = hr.closest_key
            self.assertEqual(ck(''), 'alfa')
            self.assertEqual(ck(' '), 'alfa')
            self.assertEqual(ck('alfa'), 'alfa')
            self.assertEqual(ck('bravot'), 'charlie')
            self.assertEqual(ck('charlie'), 'charlie')
            self.assertEqual(ck('kiloton'), 'lima')
            self.assertEqual(ck('oskar'), None)
            self.assertEqual(list(hr.keys()), keys)
            self.assertEqual(list(hr.values()), values)
            self.assertEqual(list(hr.keys_from('f')), keys[5:])
            hr.close()
        
    def test_stored_fields(self):
        with TempStorage("storedfields") as st:
            sf = st.create_file("test.sf")
            sfw = StoredFieldWriter(sf, ["a", "b"])
            sfw.append({"a": "hello", "b": "there"})
            sfw.append({"a": "one", "b": "two"})
            sfw.append({"a": "alfa", "b": "bravo"})
            sfw.close()
            
            sf = st.open_file("test.sf")
            sfr = StoredFieldReader(sf)
            self.assertEqual(sfr[0], {"a": "hello", "b": "there"})
            self.assertEqual(sfr[2], {"a": "alfa", "b": "bravo"})
            self.assertEqual(sfr[1], {"a": "one", "b": "two"})
            sfr.close()
        
    def test_termindex(self):
        terms = [("a", "alfa"), ("a", "bravo"), ("a", "charlie"), ("a", "delta"),
                 ("b", "able"), ("b", "baker"), ("b", "dog"), ("b", "easy")]
        st = RamStorage()
        
        tw = TermIndexWriter(st.create_file("test.trm"))
        for i, t in enumerate(terms):
            tw.add(t, (1.0, i * 1000, 1))
        tw.close()
        
        tr = TermIndexReader(st.open_file("test.trm"))
        for i, (t1, t2) in enumerate(zip(tr.keys(), terms)):
            self.assertEqual(t1, t2)
            self.assertEqual(tr.get(t1), (1.0, i * 1000, 1))
        
    

if __name__ == '__main__':
    unittest.main()



from __future__ import with_statement
import unittest

import random

from whoosh import fields
from whoosh.support.testing import TempIndex
from whoosh.util import now


class Test(unittest.TestCase):
    def test_20000_single(self):
        sc = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        with TempIndex(sc, "20000single") as ix:
            domain = ["alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
                      "golf", "hotel", "india", "juliet", "kilo", "lima"]
            
            t = now()
            for i in xrange(20000):
                w = ix.writer()
                w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
                w.commit()
            print "Write single:", now() - t
            
            t = now()
            ix.optimize()
            print "Optimize single:", now() - t

    def test_20000_buffered(self):
        from whoosh.writing import BufferedWriter
        
        sc = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        with TempIndex(sc, "20000buffered") as ix:
            domain = ["alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
                      "golf", "hotel", "india", "juliet", "kilo", "lima"]
            
            t = now()
            w = BufferedWriter(ix, limit=100, period=None)
            for i in xrange(20000):
                w.add_document(id=unicode(i),
                               text = u" ".join(random.sample(domain, 5)))
            w.close()
            print "Write buffered:", now() - t
            
            t = now()
            ix.optimize()
            print "Optimize buffered:", now() - t
            
    def test_20000_batch(self):
        sc = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        with TempIndex(sc, "20000batch") as ix:
            domain = ["alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
                      "golf", "hotel", "india", "juliet", "kilo", "lima"]
            
            t = now()
            w = ix.writer()
            for i in xrange(20000):
                w.add_document(id=unicode(i),
                               text = u" ".join(random.sample(domain, 5)))
                if not i % 100:
                    w.commit()
                    w = ix.writer()
            w.commit()
            print "Write batch:", now() - t
            
            t = now()
            ix.optimize()
            print "Optimize batch:", now() - t



if __name__ == "__main__":
    unittest.main()



from __future__ import with_statement
import unittest

from whoosh import analysis, fields, formats, index, qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.searching import Searcher
from whoosh.support.testing import TempIndex


class TestVectors(unittest.TestCase):
    def test_vector_reading(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(title = fields.TEXT,
                               content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
        
        with TempIndex(schema, "vectorreading") as ix:
            writer = ix.writer()
            writer.add_document(title=u"one",
                                content=u"This is the story of the black hole story")
            writer.commit()
            
            with ix.reader() as r:
                self.assertEqual(list(r.vector_as("frequency", 0, "content")),
                                 [(u'black', 1), (u'hole', 1), (u'story', 2)])
    
    def test_vector_merge(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(title = fields.TEXT,
                               content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
        
        with TempIndex(schema, "vectormerge") as ix:
            writer = ix.writer()
            writer.add_document(title=u"one",
                                content=u"This is the story of the black hole story")
            writer.commit()
            
            writer = ix.writer()
            writer.add_document(title=u"two",
                                content=u"You can read along in your book")
            writer.commit()
            
            with ix.searcher() as s:
                r = s.reader()
            
                docnum = s.document_number(title=u"one")
                vec = list(r.vector_as("frequency", docnum, "content"))
                self.assertEqual(vec, [(u'black', 1), (u'hole', 1), (u'story', 2)])
            
                docnum = s.document_number(title=u"two")
            
                vec = list(r.vector_as("frequency", docnum, "content"))
                self.assertEqual(vec, [(u'along', 1), (u'book', 1), (u'read', 1)])
            
    def test_vector_unicode(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
        
        with TempIndex(schema, "vectorunicode") as ix:
            writer = ix.writer()
            writer.add_document(content=u"\u1234\u2345\u3456 \u4567\u5678\u6789")
            writer.add_document(content=u"\u0123\u1234\u4567 \u4567\u5678\u6789")
            writer.commit()
            
            writer = ix.writer()
            writer.add_document(content=u"\u2345\u3456\u4567 \u789a\u789b\u789c")
            writer.add_document(content=u"\u0123\u1234\u4567 \u2345\u3456\u4567")
            writer.commit()
            
            with ix.reader() as r:
                vec = list(r.vector_as("frequency", 0, "content"))
                self.assertEqual(vec, [(u'\u3456\u4567', 1), (u'\u789a\u789b\u789c', 1)])


if __name__ == '__main__':
    unittest.main()



from __future__ import with_statement
import unittest
from random import randint, choice, sample

from whoosh import fields
from whoosh.filedb.filestore import RamStorage
from whoosh.matching import *
from whoosh.query import And, Term
from whoosh.util import make_binary_tree


class TestMatchers(unittest.TestCase):
    def _keys(self, searcher, docnums):
        return sorted([searcher.stored_fields(docnum)['key'] for docnum in docnums])
    
    def test_nullmatcher(self):
        nm = NullMatcher()
        self.assertFalse(nm.is_active())
        self.assertEqual(list(nm.all_ids()), [])
    
    def test_listmatcher(self):
        ids = [1, 2, 5, 9, 10]
        
        lm = ListMatcher(ids)
        ls = []
        while lm.is_active():
            ls.append((lm.id(), lm.score()))
            lm.next()
        self.assertEqual(ls, [(1, 1.0), (2, 1.0), (5, 1.0), (9, 1.0), (10, 1.0)])
        
        lm = ListMatcher(ids)
        self.assertEqual(list(lm.all_ids()), ids)
        
        lm = ListMatcher(ids, position=3)
        ls = []
        while lm.is_active():
            ls.append(lm.id())
            lm.next()
        self.assertEqual(ls, [9, 10])
        
        lm = ListMatcher(ids)
        for _ in xrange(3):
            lm.next()
        lm = lm.copy()
        ls = []
        while lm.is_active():
            ls.append(lm.id())
            lm.next()
        self.assertEqual(ls, [9, 10])
    
    def test_wrapper(self):
        wm = WrappingMatcher(ListMatcher([1, 2, 5, 9, 10]), boost=2.0)
        ls = []
        while wm.is_active():
            ls.append((wm.id(), wm.score()))
            wm.next()
        self.assertEqual(ls, [(1, 2.0), (2, 2.0), (5, 2.0), (9, 2.0), (10, 2.0)])
        
        ids = [1, 2, 5, 9, 10]
        wm = WrappingMatcher(ListMatcher(ids), boost=2.0)
        self.assertEqual(list(wm.all_ids()), ids)
    
    def test_filter(self):
        lm = lambda: ListMatcher(range(2, 10))
        
        fm = FilterMatcher(lm(), frozenset([3, 9]))
        self.assertEqual(list(fm.all_ids()), [3, 9])
        
        fm = FilterMatcher(lm(), frozenset([1, 5, 9, 13]))
        self.assertEqual(list(fm.all_ids()), [5, 9])
    
    def test_exclude(self):
        em = FilterMatcher(ListMatcher([1, 2, 5, 9, 10]), frozenset([2, 9]), exclude=True)
        self.assertEqual(list(em.all_ids()), [1, 5, 10])
        
        em = FilterMatcher(ListMatcher([1, 2, 5, 9, 10]), frozenset([2, 9]), exclude=True)
        self.assertEqual(list(em.all_ids()), [1, 5, 10])
        
        em = FilterMatcher(ListMatcher([1, 2, 5, 9, 10]), frozenset([2, 9]), exclude=True)
        em.next()
        em.next()
        em = em.copy()
        ls = []
        while em.is_active():
            ls.append(em.id())
            em.next()
        self.assertEqual(ls, [10])
    
    def test_simple_union(self):
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        um = UnionMatcher(lm1, lm2)
        ls = []
        while um.is_active():
            ls.append((um.id(), um.score()))
            um.next()
        self.assertEqual(ls, [(0, 1.0), (1, 1.0), (4, 2.0), (10, 1.0), (20, 2.0), (90, 1.0)])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        um = UnionMatcher(lm1, lm2)
        self.assertEqual(list(um.all_ids()), [0, 1, 4, 10, 20, 90])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        um = UnionMatcher(lm1, lm2)
        um.next()
        um.next()
        um = um.copy()
        ls = []
        while um.is_active():
            ls.append(um.id())
            um.next()
        self.assertEqual(ls, [4, 10, 20, 90])
        
    def test_simple_intersection(self):
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        im = IntersectionMatcher(lm1, lm2)
        ls = []
        while im.is_active():
            ls.append((im.id(), im.score()))
            im.next()
        self.assertEqual(ls, [(4, 2.0), (20, 2.0)])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        im = IntersectionMatcher(lm1, lm2)
        self.assertEqual(list(im.all_ids()), [4, 20])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        im = IntersectionMatcher(lm1, lm2)
        im.next()
        im.next()
        im = im.copy()
        ls = []
        while im.is_active():
            ls.append(im.id())
            im.next()
        self.assertEqual(ls, [])
    
    def test_andnot(self):
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        anm = AndNotMatcher(lm1, lm2)
        ls = []
        while anm.is_active():
            ls.append((anm.id(), anm.score()))
            anm.next()
        self.assertEqual(ls, [(1, 1.0), (10, 1.0), (90, 1.0)])
        
        echo_lm = ListMatcher([0, 1, 2, 3, 4])
        bravo_lm = ListMatcher([0, 1])
        anm = AndNotMatcher(echo_lm, bravo_lm)
        self.assertEqual(list(anm.all_ids()), [2, 3, 4])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        anm = AndNotMatcher(lm1, lm2)
        self.assertEqual(list(anm.all_ids()), [1, 10, 90])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        anm = AndNotMatcher(lm1, lm2)
        anm.next()
        anm.next()
        anm = anm.copy()
        ls = []
        while anm.is_active():
            ls.append(anm.id())
            anm.next()
        self.assertEqual(ls, [90])
    
    def test_require(self):
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        rm = RequireMatcher(lm1, lm2)
        ls = []
        while rm.is_active():
            ls.append((rm.id(), rm.score()))
            rm.next()
        self.assertEqual(ls, [(4, 1.0), (20, 1.0)])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        rm = RequireMatcher(lm1, lm2)
        self.assertEqual(list(rm.all_ids()), [4, 20])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        rm = RequireMatcher(lm1, lm2)
        rm.next()
        rm.next()
        rm = rm.copy()
        ls = []
        while rm.is_active():
            ls.append(rm.id())
            rm.next()
        self.assertEqual(ls, [])
    
    def test_andmaybe(self):
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        amm = AndMaybeMatcher(lm1, lm2)
        ls = []
        while amm.is_active():
            ls.append((amm.id(), amm.score()))
            amm.next()
        self.assertEqual(ls, [(1, 1.0), (4, 2.0), (10, 1.0), (20, 2.0), (90, 1.0)])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        amm = AndMaybeMatcher(lm1, lm2)
        self.assertEqual(list(amm.all_ids()), [1, 4, 10, 20, 90])
        
        lm1 = ListMatcher([1, 4, 10, 20, 90])
        lm2 = ListMatcher([0, 4, 20])
        amm = AndMaybeMatcher(lm1, lm2)
        amm.next()
        amm.next()
        amm = amm.copy()
        ls = []
        while amm.is_active():
            ls.append(amm.id())
            amm.next()
        self.assertEqual(ls, [10, 20, 90])
    
    def test_intersection(self):
        schema = fields.Schema(key = fields.ID(stored=True), value = fields.TEXT(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(key=u"a", value=u"alpha bravo charlie delta")
        w.add_document(key=u"b", value=u"echo foxtrot alpha bravo")
        w.add_document(key=u"c", value=u"charlie delta golf hotel")
        w.commit()
        
        w = ix.writer()
        w.add_document(key=u"d", value=u"india alpha bravo charlie")
        w.add_document(key=u"e", value=u"delta bravo india bravo")
        w.commit()
        
        with ix.searcher() as s:
            q = And([Term("value", u"bravo"), Term("value", u"delta")])
            m = q.matcher(s)
            self.assertEqual(self._keys(s, m.all_ids()), ["a", "e"])
            
            q = And([Term("value", u"bravo"), Term("value", u"alpha")])
            m = q.matcher(s)
            self.assertEqual(self._keys(s, m.all_ids()), ["a", "b", "d"])
        
    def test_random_intersections(self):
        domain = [u"alpha", u"bravo", u"charlie", u"delta", u"echo",
                  u"foxtrot", u"golf", u"hotel", u"india", u"juliet", u"kilo",
                  u"lima", u"mike"]
        segments = 5
        docsperseg = 50
        fieldlimits = (3, 10)
        documents = []
        
        schema = fields.Schema(key=fields.STORED, value=fields.TEXT(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        # Create docsperseg * segments documents containing random words from
        # the domain list. Add the documents to the index, but also keep them
        # in the "documents" list for the sanity check
        for i in xrange(segments):
            w = ix.writer()
            for j in xrange(docsperseg):
                docnum = i * docsperseg + j
                # Create a string of random words
                doc = u" ".join(choice(domain)
                                for _ in xrange(randint(*fieldlimits)))
                # Add the string to the index
                w.add_document(key=docnum, value=doc)
                # Add a (docnum, string) tuple to the documents list
                documents.append((docnum, doc))
            w.commit()
        self.assertNotEqual(len(ix._segments()), 1)
        
        testcount = 20
        testlimits = (2, 5)
        
        with ix.searcher() as s:
            for i in xrange(s.doc_count_all()):
                self.assertNotEqual(s.stored_fields(i).get("key"), None)
            
            for _ in xrange(testcount):
                # Create a random list of words and manually do an intersection of
                # items in "documents" that contain the words ("target").
                words = sample(domain, randint(*testlimits))
                target = []
                for docnum, doc in documents:
                    if all((doc.find(w) > -1) for w in words):
                        target.append(docnum)
                target.sort()
                
                # Create a query from the list of words and get two matchers from
                # it.
                q = And([Term("value", w) for w in words])
                m1 = q.matcher(s)
                m2 = q.matcher(s)
                
                # Try getting the list of IDs from all_ids()
                ids1 = list(m1.all_ids())
                
                # Try getting the list of IDs using id()/next()
                ids2 = []
                while m2.is_active():
                    ids2.append(m2.id())
                    m2.next()
                
                # Check that the two methods return the same list
                self.assertEqual(ids1, ids2)
                
                # Check that the IDs match the ones we manually calculated
                keys = self._keys(s, ids1)
                self.assertEqual(keys, target)

    def test_union(self):
        s1 = ListMatcher([1, 2, 3, 4, 5, 6, 7, 8])
        s2 = ListMatcher([2, 4, 8, 10, 20, 30])
        s3 = ListMatcher([10, 100, 200])
        target = [1, 2, 3, 4, 5, 6, 7, 8, 10, 20, 30, 100, 200]
        um = UnionMatcher(s1, UnionMatcher(s2, s3))
        self.assertEqual(target, list(um.all_ids()))
        
    def test_union_scores(self):
        s1 = ListMatcher([1, 2, 3])
        s2 = ListMatcher([2, 4, 8])
        s3 = ListMatcher([2, 3, 8])
        target = [(1, 1.0), (2, 3.0), (3, 2.0), (4, 1.0), (8, 2.0)]
        um = UnionMatcher(s1, UnionMatcher(s2, s3))
        result = []
        while um.is_active():
            result.append((um.id(), um.score()))
            um.next()
        self.assertEqual(target, result)

    def test_random_union(self):
        testcount = 100
        rangelimits = (2, 10)
        clauselimits = (2, 10)
        
        vals = range(100)
        
        for _ in xrange(testcount):
            target = set()
            matchers = []
            for _ in xrange(randint(*clauselimits)):
                nums = sample(vals, randint(*rangelimits))
                target = target.union(nums)
                matchers.append(ListMatcher(sorted(nums)))
            target = sorted(target)
            um = make_binary_tree(UnionMatcher, matchers)
            self.assertEqual(list(um.all_ids()), target)

    def test_inverse(self):
        s = ListMatcher([1, 5, 10, 11, 13])
        inv = InverseMatcher(s, 15)
        ids = []
        while inv.is_active():
            ids.append(inv.id())
            inv.next()
        self.assertEqual(ids, [0, 2, 3, 4, 6, 7, 8, 9, 12, 14])
        
    def test_inverse_skip(self):
        s = ListMatcher([1, 5, 10, 11, 13])
        inv = InverseMatcher(s, 15)
        inv.skip_to(8)
        
        ids = []
        while inv.is_active():
            ids.append(inv.id())
            inv.next()
        self.assertEqual([8, 9, 12, 14], ids)

    def test_empty_andnot(self):
        pos = NullMatcher()
        neg = NullMatcher()
        anm = AndNotMatcher(pos, neg)
        self.assertFalse(anm.is_active())
        self.assertEqual(list(anm.all_ids()), [])
        
        pos = ListMatcher([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        neg = NullMatcher()
        ans = AndNotMatcher(pos, neg)
        ids = list(ans.all_ids())
        self.assertEqual(ids, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_random_andnot(self):
        testcount = 100
        rangesize = 100
        
        rng = range(rangesize)
        
        for _ in xrange(testcount):
            negs = sorted(sample(rng, randint(0, rangesize-1)))
            negset = frozenset(negs)
            matched = [n for n in rng if n not in negset]
            
            pos = ListMatcher(rng)
            neg = ListMatcher(negs)
            
            anm = AndNotMatcher(pos, neg)
            ids = list(anm.all_ids())
            self.assertEqual(ids, matched)


if __name__ == '__main__':
    unittest.main()





# coding=utf-8

import unittest

from whoosh import analysis, fields, qparser, query


class TestQueryParser(unittest.TestCase):
    def test_fields(self):
        s = fields.Schema(content=fields.TEXT, title=fields.TEXT, id=fields.ID)
        
        qp = qparser.QueryParser("content", s)
        q = qp.parse(u"test")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "content")
        self.assertEqual(q.text, "test")
        
        mq = qparser.MultifieldParser(("title", "content"), s)
        q = mq.parse(u"test")
        self.assertEqual(q.__class__, query.Or)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[0].fieldname, "title")
        self.assertEqual(q[1].fieldname, "content")
        self.assertEqual(q[0].text, "test")
        self.assertEqual(q[1].text, "test")
        
        q = mq.parse(u"title:test")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "title")
        self.assertEqual(q.text, "test")
    
    def test_colonspace(self):
        s = fields.Schema(content=fields.TEXT, url=fields.ID)
        
        qp = qparser.QueryParser("content", s)
        q = qp.parse(u"url:test")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "url")
        self.assertEqual(q.text, "test")
        
        q = qp.parse(u"url: test")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[0].fieldname, "content")
        self.assertEqual(q[1].fieldname, "content")
        self.assertEqual(q[0].text, "url")
        self.assertEqual(q[1].text, "test")
        
        q = qp.parse(u"url:")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "content")
        self.assertEqual(q.text, "url")
    
    def test_andor(self):
        qp  = qparser.QueryParser("a", None)
        q = qp.parse("a AND b OR c AND d OR e AND f")
        self.assertEqual(unicode(q), u"((a:a AND a:b) OR (a:c AND a:d) OR (a:e AND a:f))")
        
        q = qp.parse("aORb")
        self.assertEqual(q, query.Term("a", "aORb"))
        
        q = qp.parse("aOR b")
        self.assertEqual(q, query.And([query.Term("a", "aOR"), query.Term("a", "b")]))
        
        q = qp.parse("a ORb")
        self.assertEqual(q, query.And([query.Term("a", "a"), query.Term("a", "ORb")]))
        
        self.assertEqual(qp.parse("OR"), query.Term("a", "OR"))
    
    def test_andnot(self):
        qp = qparser.QueryParser("content", None)
        q = qp.parse(u"this ANDNOT that")
        self.assertEqual(q.__class__, query.AndNot)
        self.assertEqual(q.a.__class__, query.Term)
        self.assertEqual(q.b.__class__, query.Term)
        self.assertEqual(q.a.text, "this")
        self.assertEqual(q.b.text, "that")
        
        q = qp.parse(u"foo ANDNOT bar baz")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.AndNot)
        self.assertEqual(q[1].__class__, query.Term)
        
        q = qp.parse(u"foo fie ANDNOT bar baz")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.AndNot)
        self.assertEqual(q[2].__class__, query.Term)
        
        q = qp.parse(u"a AND b ANDNOT c")
        self.assertEqual(q.__class__, query.AndNot)
        self.assertEqual(unicode(q), "((content:a AND content:b) ANDNOT content:c)")
    
    def test_boost(self):
        qp = qparser.QueryParser("content", None)
        q = qp.parse(u"this^3 fn:that^0.5 5.67 hi^5x")
        self.assertEqual(q[0].boost, 3.0)
        self.assertEqual(q[1].boost, 0.5)
        self.assertEqual(q[1].fieldname, "fn")
        self.assertEqual(q[2].text, "5.67")
        self.assertEqual(q[3].text, "hi^5x")
        
        q = qp.parse("alfa (bravo OR charlie)^2.5 ^3")
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].boost, 1.0)
        self.assertEqual(q[1].boost, 2.5)
        self.assertEqual(q[2].text, "^3")
        
    def test_wildcard1(self):
        qp = qparser.QueryParser("content", None)
        q = qp.parse(u"hello *the?e* ?star*s? test")
        self.assertEqual(len(q), 4)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].text, "hello")
        self.assertEqual(q[1].__class__, query.Wildcard)
        self.assertEqual(q[1].text, "*the?e*")
        self.assertEqual(q[2].__class__, query.Wildcard)
        self.assertEqual(q[2].text, "?star*s?")
        self.assertEqual(q[3].__class__, query.Term)
        self.assertEqual(q[3].text, "test")
        
    def test_wildcard2(self):
        qp = qparser.QueryParser("content", None)
        q = qp.parse(u"*the?e*")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, "*the?e*")
        
    def test_parse_fieldname_underscores(self):
        s = fields.Schema(my_name=fields.ID(stored=True), my_value=fields.TEXT)
        qp = qparser.QueryParser("my_value", schema=s)
        q = qp.parse(u"my_name:Green")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "my_name")
        self.assertEqual(q.text, "Green")
    
    def test_endstar(self):
        qp = qparser.QueryParser("text", None)
        q = qp.parse(u"word*")
        self.assertEqual(q.__class__, query.Prefix)
        self.assertEqual(q.text, "word")
        
        q = qp.parse(u"first* second")
        self.assertEqual(q[0].__class__, query.Prefix)
        self.assertEqual(q[0].text, "first")
    
    def test_singlequotes(self):
        qp = qparser.QueryParser("text", None)
        q = qp.parse("hell's hot 'i stab at thee'")
        self.assertEqual(q.__class__.__name__, 'And')
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[2].__class__, query.Term)
        self.assertEqual(q[0].text, "hell's")
        self.assertEqual(q[1].text, "hot")
        self.assertEqual(q[2].text, "i stab at thee")
        
        q = qp.parse("alfa zulu:'bravo charlie' delta")
        self.assertEqual(q.__class__.__name__, 'And')
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[2].__class__, query.Term)
        self.assertEqual((q[0].fieldname, q[0].text), ("text", "alfa"))
        self.assertEqual((q[1].fieldname, q[1].text), ("zulu", "bravo charlie"))
        self.assertEqual((q[2].fieldname, q[2].text), ("text", "delta"))
        
        q = qp.parse("The rest 'is silence")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 4)
        self.assertEqual([t.text for t in q.subqueries],
                         ["The", "rest", "'is" ,"silence"])
        
        q = qp.parse("I don't like W's stupid face")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 6)
        self.assertEqual([t.text for t in q.subqueries],
                         ["I", "don't", "like" ,"W's", "stupid", "face"])
        
        q = qp.parse("I forgot the drinkin' in '98")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 6)
        self.assertEqual([t.text for t in q.subqueries],
                         ["I", "forgot", "the" ,"drinkin'", "in", "'98"])
    
#    def test_escaping(self):
#        qp = qparser.QueryParser("text", None)
#        
#        q = qp.parse(r'big\small')
#        self.assertEqual(q.__class__, query.Term, q)
#        self.assertEqual(q.text, "bigsmall")
#        
#        q = qp.parse(r'big\\small')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, r'big\small')
#        
#        q = qp.parse(r'http\:example')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.fieldname, "text")
#        self.assertEqual(q.text, "http:example")
#        
#        q = qp.parse(r'hello\ there')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "hello there")
#        
#        q = qp.parse(r'\[start\ TO\ end\]')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "[start TO end]")
#    
#        schema = fields.Schema(text=fields.TEXT)
#        qp = qparser.QueryParser("text", None)
#        q = qp.parse(r"http\:\/\/www\.example\.com")
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "http://www.example.com")
#        
#        q = qp.parse(u"\\\\")
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "\\")
    
#    def test_escaping_wildcards(self):
#        qp = qparser.QueryParser("text", None)
#        
#        q = qp.parse(u"a*b*c?d")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, "a*b*c?d")
#        
#        q = qp.parse(u"a*b\\*c?d")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, "a*b*c?d")
#        
#        q = qp.parse(u"a*b\\\\*c?d")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, u'a*b\\*c?d')
#        
#        q = qp.parse(u"ab*")
#        self.assertEqual(q.__class__, query.Prefix)
#        self.assertEqual(q.text, u"ab")
#        
#        q = qp.parse(u"ab\\\\*")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, u"ab\\*")
        
    def test_phrase(self):
        qp = qparser.QueryParser("content", None)
        q = qp.parse('"alfa bravo" "charlie delta echo"^2.2 test:"foxtrot golf"')
        self.assertEqual(q[0].__class__, query.Phrase)
        self.assertEqual(q[0].words, ["alfa", "bravo"])
        self.assertEqual(q[1].__class__, query.Phrase)
        self.assertEqual(q[1].words, ["charlie", "delta", "echo"])
        self.assertEqual(q[1].boost, 2.2)
        self.assertEqual(q[2].__class__, query.Phrase)
        self.assertEqual(q[2].words, ["foxtrot", "golf"])
        self.assertEqual(q[2].fieldname, "test")
        
    def test_weird_characters(self):
        qp = qparser.QueryParser("content", None)
        q = qp.parse(u".abcd@gmail.com")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, u".abcd@gmail.com")
        q = qp.parse(u"r*")
        self.assertEqual(q.__class__, query.Prefix)
        self.assertEqual(q.text, u"r")
        q = qp.parse(u".")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, u".")
        q = qp.parse(u"?")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"?")
        
    def test_euro_chars(self):
        schema = fields.Schema(text=fields.TEXT)
        qp = qparser.QueryParser("text", schema)
        q = qp.parse(u"straße")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, u"straße")
    
    def test_star(self):
        schema = fields.Schema(text = fields.TEXT(stored=True))
        qp = qparser.QueryParser("text", schema)
        q = qp.parse(u"*")
        self.assertEqual(q.__class__, query.Every)
        
        q = qp.parse(u"*h?ll*")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"*h?ll*")
        
        q = qp.parse(u"h?pe")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"h?pe")
        
        q = qp.parse(u"*? blah")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Wildcard)
        self.assertEqual(q[0].text, u"*?")
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].text, u"blah")
        
        q = qp.parse(u"*ending")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"*ending")
        
        q = qp.parse(u"*q")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"*q")

    def test_range(self):
        schema = fields.Schema(name=fields.ID(stored=True), text = fields.TEXT(stored=True))
        qp = qparser.QueryParser("text", schema)
        
        q = qp.parse(u"[alfa to bravo}")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "alfa")
        self.assertEqual(q.end, "bravo")
        self.assertEqual(q.startexcl, False)
        self.assertEqual(q.endexcl, True)
        
        q = qp.parse(u"['hello there' to 'what ever']")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "hello there")
        self.assertEqual(q.end, "what ever")
        self.assertEqual(q.startexcl, False)
        self.assertEqual(q.endexcl, False)
        
        q = qp.parse(u"name:{'to' to 'b'}")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "to")
        self.assertEqual(q.end, "b")
        self.assertEqual(q.startexcl, True)
        self.assertEqual(q.endexcl, True)
        
        q = qp.parse(u"name:{'a' to 'to']")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "a")
        self.assertEqual(q.end, "to")
        self.assertEqual(q.startexcl, True)
        self.assertEqual(q.endexcl, False)
        
        q = qp.parse(u"name:[a to to]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "a")
        self.assertEqual(q.end, "to")
        
        q = qp.parse(u"name:[to to b]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "to")
        self.assertEqual(q.end, "b")
        
        q = qp.parse(u"[alfa to alfa]")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, "alfa")
        
        q = qp.parse(u"Ind* AND name:[d TO]")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Prefix)
        self.assertEqual(q[1].__class__, query.TermRange)
        self.assertEqual(q[0].text, "ind")
        self.assertEqual(q[1].start, "d")
        self.assertEqual(q[1].fieldname, "name")
        
        q = qp.parse(u"name:[d TO]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "d")
        self.assertEqual(q.fieldname, "name")
    
    def test_numeric_range(self):
        schema = fields.Schema(id=fields.STORED, number=fields.NUMERIC)
        qp = qparser.QueryParser("number", schema)
        
        teststart = 40
        testend = 100
        
        q = qp.parse("[%s to *]" % teststart)
        self.assertEqual(q, query.NullQuery)
        
        q = qp.parse("[%s to]" % teststart)
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, teststart)
        self.assertEqual(q.end, None)
        
        q = qp.parse("[to %s]" % testend)
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, None)
        self.assertEqual(q.end, testend)
        
        q = qp.parse("[%s to %s]" % (teststart, testend))
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, teststart)
        self.assertEqual(q.end, testend)
        
    def test_regressions(self):
        qp = qparser.QueryParser("f", None)
        
        # From 0.3.18, these used to require escaping. Mostly good for
        # regression testing.
        self.assertEqual(qp.parse(u"re-inker"), query.Term("f", u"re-inker"))
        self.assertEqual(qp.parse(u"0.7 wire"),
                         query.And([query.Term("f", u"0.7"), query.Term("f", u"wire")]))
        self.assertEqual(qp.parse(u"daler-rowney pearl 'bell bronze'"),
                         query.And([query.Term("f", u"daler-rowney"),
                                    query.Term("f", u"pearl"),
                                    query.Term("f", u"bell bronze")]))
        
        q = qp.parse(u'22" BX')
        self.assertEqual(q, query.And([query.Term("f", u'22"'), query.Term("f", "BX")]))
        
    def test_empty_ranges(self):
        schema = fields.Schema(name=fields.TEXT, num=fields.NUMERIC,
                               date=fields.DATETIME)
        qp = qparser.QueryParser("text", schema)
        
        for fname in ("name", "date"):
            q = qp.parse("%s:[to]" % fname)
            self.assertEqual(q.__class__, query.Every)
    
    def test_empty_numeric_range(self):
        schema = fields.Schema(id=fields.ID, num=fields.NUMERIC)
        qp = qparser.QueryParser("num", schema)
        q = qp.parse("num:[to]")
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, None)
        self.assertEqual(q.end, None)
    
    def test_nonexistant_fieldnames(self):
        # Need an analyzer that won't mangle a URL
        a = analysis.SimpleAnalyzer("\\S+")
        schema = fields.Schema(id=fields.ID, text=fields.TEXT(analyzer=a))
        
        qp = qparser.QueryParser("text", schema)
        q = qp.parse(u"id:/code http://localhost/")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].fieldname, "id")
        self.assertEqual(q[0].text, "/code")
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].fieldname, "text")
        self.assertEqual(q[1].text, "http://localhost/")
    
    def test_stopped(self):
        schema = fields.Schema(text = fields.TEXT)
        qp = qparser.QueryParser("text", schema)
        q = qp.parse(u"a b")
        self.assertEqual(q, query.NullQuery)
        
    def test_analyzing_terms(self):
        schema = fields.Schema(text=fields.TEXT(analyzer=analysis.StemmingAnalyzer()))
        qp = qparser.QueryParser("text", schema)
        q = qp.parse(u"Indexed!")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, "index")
        
    def test_simple(self):
        parser = qparser.SimpleParser("x", None)
        q = parser.parse(u"alfa bravo charlie delta")
        self.assertEqual(unicode(q), u"(x:alfa OR x:bravo OR x:charlie OR x:delta)")
        
        q = parser.parse(u"alfa +bravo charlie delta")
        self.assertEqual(unicode(q), u"(x:bravo ANDMAYBE (x:alfa OR x:charlie OR x:delta))")
        
        q = parser.parse(u"alfa +bravo -charlie delta")
        self.assertEqual(unicode(q), u"((x:bravo ANDMAYBE (x:alfa OR x:delta)) ANDNOT x:charlie)")
        
        q = parser.parse(u"- alfa +bravo + delta")
        self.assertEqual(unicode(q), u"((x:bravo AND x:delta) ANDNOT x:alfa)")
    
    def test_dismax(self):
        parser = qparser.DisMaxParser({"body": 0.8, "title": 2.5}, None)
        q = parser.parse(u"alfa bravo charlie")
        self.assertEqual(unicode(q), u"(DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:bravo^0.8 title:bravo^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5))")
        
        q = parser.parse(u"alfa +bravo charlie")
        self.assertEqual(unicode(q), u"(DisMax(body:bravo^0.8 title:bravo^2.5) ANDMAYBE (DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5)))")
        
        q = parser.parse(u"alfa -bravo charlie")
        self.assertEqual(unicode(q), u"((DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5)) ANDNOT DisMax(body:bravo^0.8 title:bravo^2.5))")
        
        q = parser.parse(u"alfa -bravo +charlie")
        self.assertEqual(unicode(q), u"((DisMax(body:charlie^0.8 title:charlie^2.5) ANDMAYBE DisMax(body:alfa^0.8 title:alfa^2.5)) ANDNOT DisMax(body:bravo^0.8 title:bravo^2.5))")
    
    def test_many_clauses(self):
        parser = qparser.QueryParser("content", None)
        qs = "1" + (" OR 1" * 1000)
        parser.parse(qs)
        
        parser = qparser.QueryParser("content", None)
        parser.parse(qs)
    
    def test_roundtrip(self):
        parser = qparser.QueryParser("a", None)
        q = parser.parse(u"a OR ((b AND c AND d AND e) OR f OR g) ANDNOT h")
        self.assertEqual(unicode(q), u"((a:a OR (a:b AND a:c AND a:d AND a:e) OR a:f OR a:g) ANDNOT a:h)")
        
    def test_ngrams(self):
        schema = fields.Schema(grams=fields.NGRAM)
        parser = qparser.QueryParser('grams', schema)
        parser.remove_plugin_class(qparser.WhitespacePlugin)
        
        q = parser.parse(u"Hello There")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 8)
        self.assertEqual([sq.text for sq in q], ["hell", "ello", "llo ", "lo t", "o th", " the", "ther", "here"])
        
    def test_ngramwords(self):
        schema = fields.Schema(grams=fields.NGRAMWORDS(queryor=True))
        parser = qparser.QueryParser('grams', schema)
        
        q = parser.parse(u"Hello Tom")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Or)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[0][0].text, "hell")
        self.assertEqual(q[0][1].text, "ello")
        self.assertEqual(q[1].text, "tom")

    def test_multitoken_words(self):
        textfield = fields.TEXT()
        textfield.multitoken_query = "or"
        schema = fields.Schema(text=textfield)
        parser = qparser.QueryParser('text', schema)
        qstring = u"chaw-bacon"
        
        texts = list(schema["text"].process_text(qstring))
        self.assertEqual(texts, ["chaw", "bacon"])
        
        q = parser.parse(qstring)
        self.assertEqual(q.__class__, query.Or)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].text, "chaw")
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].text, "bacon")
    
    def test_operators(self):
        qp = qparser.QueryParser("f", None)
        
        q = qp.parse("a AND b OR c AND d")
        self.assertEqual(unicode(q), "((f:a AND f:b) OR (f:c AND f:d))")
        
        q = qp.parse("a OR b OR c OR d")
        self.assertEqual(unicode(q), "(f:a OR f:b OR f:c OR f:d)")
        
        q = qp.parse("a ANDMAYBE b ANDNOT c REQUIRE d")
        self.assertEqual(unicode(q), "((f:a ANDMAYBE (f:b ANDNOT f:c)) REQUIRE f:d)")
    
    def test_associativity(self):
        left_andmaybe = (qparser.InfixOperator("ANDMAYBE", qparser.AndMaybeGroup, True), 0)
        right_andmaybe = (qparser.InfixOperator("ANDMAYBE", qparser.AndMaybeGroup, False), 0)
        not_ = (qparser.PrefixOperator("NOT", qparser.NotGroup), 0)
        
        def make_parser(*ops):
            parser = qparser.QueryParser("f", None)
            parser.replace_plugin(qparser.CompoundsPlugin(ops, clean=True))
            return parser
        
        p = make_parser(left_andmaybe)
        q = p.parse("a ANDMAYBE b ANDMAYBE c ANDMAYBE d")
        self.assertEqual(unicode(q), "(((f:a ANDMAYBE f:b) ANDMAYBE f:c) ANDMAYBE f:d)")
        
        p = make_parser(right_andmaybe)
        q = p.parse("a ANDMAYBE b ANDMAYBE c ANDMAYBE d")
        self.assertEqual(unicode(q), "(f:a ANDMAYBE (f:b ANDMAYBE (f:c ANDMAYBE f:d)))")
        
        p = make_parser(not_)
        q = p.parse("a NOT b NOT c NOT d", normalize=False)
        self.assertEqual(unicode(q), "(f:a AND NOT f:b AND NOT f:c AND NOT f:d)")
        
        p = make_parser(left_andmaybe)
        q = p.parse("(a ANDMAYBE b) ANDMAYBE (c ANDMAYBE d)")
        self.assertEqual(unicode(q), "((f:a ANDMAYBE f:b) ANDMAYBE (f:c ANDMAYBE f:d))")
        
        p = make_parser(right_andmaybe)
        q = p.parse("(a ANDMAYBE b) ANDMAYBE (c ANDMAYBE d)")
        self.assertEqual(unicode(q), "((f:a ANDMAYBE f:b) ANDMAYBE (f:c ANDMAYBE f:d))")
        
    def test_not_assoc(self):
        qp = qparser.QueryParser("text", None)
        q = qp.parse(u"a AND NOT b OR c")
        self.assertEqual(unicode(q), "((text:a AND NOT text:b) OR text:c)")
        
        qp = qparser.QueryParser("text", None)
        q = qp.parse(u"a NOT (b OR c)")
        self.assertEqual(unicode(q), "(text:a AND NOT (text:b OR text:c))")
        
    def test_fieldname_space(self):
        qp = qparser.QueryParser("a", None)
        q = qp.parse("Man Ray: a retrospective")
        self.assertEqual(unicode(q), "(a:Man AND a:Ray: AND a:a AND a:retrospective)")
        
    def test_fieldname_fieldname(self):
        qp = qparser.QueryParser("a", None)
        q = qp.parse("a:b:")
        self.assertEqual(q, query.Term("a", u"b:"))
        
    def test_paren_fieldname(self):
        schema = fields.Schema(kind=fields.ID, content=fields.TEXT) 
    
        qp = qparser.QueryParser("content", schema)
        q = qp.parse(u"(kind:1d565 OR kind:7c584) AND (stuff)")
        self.assertEqual(unicode(q), "((kind:1d565 OR kind:7c584) AND content:stuff)")

        q = qp.parse(u"kind:(1d565 OR 7c584) AND (stuff)")
        self.assertEqual(unicode(q), "((kind:1d565 OR kind:7c584) AND content:stuff)")


if __name__ == '__main__':
    unittest.main()



"""This module implements the Lovins stemming algorithm. Use the ``stem()``
function::

    stemmed_word = stem(word)
"""

from collections import defaultdict


# Conditions

def A(base):
    # A   No restrictions on stem
    return True


def B(base):
    # B  Minimum stem length = 3
    return len(base) > 2


def C(base):
    # C  Minimum stem length = 4
    return len(base) > 3


def D(base):
    # D  Minimum stem length = 5
    return len(base) > 4


def E(base):
    # E  Do not remove ending after e
    return base[-1] != "e"


def F(base):
    # F  Minimum stem length = 3 and do not remove ending after e
    return len(base) > 2 and base[-1] != "e"


def G(base):
    # G  Minimum stem length = 3 and remove ending only after f
    return len(base) > 2 and base[-1] == "f"


def H(base):
    # H  Remove ending only after t or ll
    c1, c2 = base[-2:]
    return c2 == "t" or (c2 == "l" and c1 == "l")


def I(base):
    # I  Do not remove ending after o or e
    c = base[-1]
    return c != "o" and c != "e"


def J(base):
    # J  Do not remove ending after a or e
    c = base[-1]
    return c != "a" and c != "e"


def K(base):
    # K  Minimum stem length = 3 and remove ending only after l, i or u*e
    c = base[-1]
    cc = base[-3]
    return len(base) > 2 and (c == "l" or c == "i" or (c == "e" and cc == "u"))


def L(base):
    # L  Do not remove ending after u, x or s, unless s follows o
    c1, c2 = base[-2:]
    return c2 != "u" and c2 != "x" and (c2 != "s" or c1 == "o")


def M(base):
    # M  Do not remove ending after a, c, e or m
    c = base[-1]
    return c != "a" and c != "c" and c != "e" and c != "m"


def N(base):
    # N  Minimum stem length = 4 after s**, elsewhere = 3
    return len(base) > 3 or (len(base) == 3 and base[-1] != "s")


def O(base):
    # O  Remove ending only after l or i
    c = base[-1]
    return c == "l" or c == "i"


def P(base):
    # P  Do not remove ending after c
    return base[-1] != "c"


def Q(base):
    # Q  Minimum stem length = 3 and do not remove ending after l or n
    c = base[-1]
    return len(base) > 2 and (c != "l" and c != "n")


def R(base):
    # R  Remove ending only after n or r
    c = base[-1]
    return c == "n" or c == "r"


def S(base):
    # S  Remove ending only after dr or t, unless t follows t
    l2 = base[-2]
    return l2 == "rd" or (base[-1] == "t" and l2 != "tt")


def T(base):
    # T  Remove ending only after s or t, unless t follows o
    c1, c2 = base[-2:]
    return c2 == "s" or (c2 == "t" and c1 != "o")


def U(base):
    # U  Remove ending only after l, m, n or r
    c = base[-1]
    return c == "l" or c == "m" or c == "n" or c == "r"


def V(base):
    # V  Remove ending only after c
    return base[-1] == "c"


def W(base):
    # W  Do not remove ending after s or u
    c = base[-1]
    return c != "s" and c != "u"


def X(base):
    # X  Remove ending only after l, i or u*e
    c = base[-1]
    cc = base[-3]
    return c == "l" or c == "i" or (c == "e" and cc == "u")


def Y(base):
    # Y  Remove ending only after in
    return base[-2:] == "in"


def Z(base):
    # Z  Do not remove ending after f
    return base[-1] != "f"


def a(base):
    # a  Remove ending only after d, f, ph, th, l, er, or, es or t
    c = base[-1]
    l2 = base[-2:]
    return (c == "d" or c == "f" or l2 == "ph" or l2 == "th" or c == "l"
            or l2 == "er" or l2 == "or" or l2 == "es" or c == "t")


def b(base):
    # b  Minimum stem length = 3 and do not remove ending after met or ryst
    return len(base) > 2 and not (base.endswith("met")
                                  or base.endswith("ryst"))


def c(base):
    # c  Remove ending only after l
    return base[-1] == "l"


# Endings

m = [None] * 12

m[11] = dict((
        ("alistically", B),
        ("arizability", A),
        ("izationally", B)))
m[10] = dict((
        ("antialness", A),
        ("arisations", A),
        ("arizations", A),
        ("entialness", A)))
m[9] = dict((
        ("allically", C),
        ("antaneous", A),
        ("antiality", A),
        ("arisation", A),
        ("arization", A),
        ("ationally", B),
        ("ativeness", A),
        ("eableness", E),
        ("entations", A),
        ("entiality", A),
        ("entialize", A),
        ("entiation", A),
        ("ionalness", A),
        ("istically", A),
        ("itousness", A),
        ("izability", A),
        ("izational", A)))
m[8] = dict((
        ("ableness", A),
        ("arizable", A),
        ("entation", A),
        ("entially", A),
        ("eousness", A),
        ("ibleness", A),
        ("icalness", A),
        ("ionalism", A),
        ("ionality", A),
        ("ionalize", A),
        ("iousness", A),
        ("izations", A),
        ("lessness", A)))
m[7] = dict((
        ("ability", A),
        ("aically", A),
        ("alistic", B),
        ("alities", A),
        ("ariness", E),
        ("aristic", A),
        ("arizing", A),
        ("ateness", A),
        ("atingly", A),
        ("ational", B),
        ("atively", A),
        ("ativism", A),
        ("elihood", E),
        ("encible", A),
        ("entally", A),
        ("entials", A),
        ("entiate", A),
        ("entness", A),
        ("fulness", A),
        ("ibility", A),
        ("icalism", A),
        ("icalist", A),
        ("icality", A),
        ("icalize", A),
        ("ication", G),
        ("icianry", A),
        ("ination", A),
        ("ingness", A),
        ("ionally", A),
        ("isation", A),
        ("ishness", A),
        ("istical", A),
        ("iteness", A),
        ("iveness", A),
        ("ivistic", A),
        ("ivities", A),
        ("ization", F),
        ("izement", A),
        ("oidally", A),
        ("ousness", A)))
m[6] = dict((
        ("aceous", A),
        ("acious", B),
        ("action", G),
        ("alness", A),
        ("ancial", A),
        ("ancies", A),
        ("ancing", B),
        ("ariser", A),
        ("arized", A),
        ("arizer", A),
        ("atable", A),
        ("ations", B),
        ("atives", A),
        ("eature", Z),
        ("efully", A),
        ("encies", A),
        ("encing", A),
        ("ential", A),
        ("enting", C),
        ("entist", A),
        ("eously", A),
        ("ialist", A),
        ("iality", A),
        ("ialize", A),
        ("ically", A),
        ("icance", A),
        ("icians", A),
        ("icists", A),
        ("ifully", A),
        ("ionals", A),
        ("ionate", D),
        ("ioning", A),
        ("ionist", A),
        ("iously", A),
        ("istics", A),
        ("izable", E),
        ("lessly", A),
        ("nesses", A),
        ("oidism", A)))
m[5] = dict((
        ("acies", A),
        ("acity", A),
        ("aging", B),
        ("aical", A),
        ("alist", A),
        ("alism", B),
        ("ality", A),
        ("alize", A),
        ("allic", b),
        ("anced", B),
        ("ances", B),
        ("antic", C),
        ("arial", A),
        ("aries", A),
        ("arily", A),
        ("arity", B),
        ("arize", A),
        ("aroid", A),
        ("ately", A),
        ("ating", I),
        ("ation", B),
        ("ative", A),
        ("ators", A),
        ("atory", A),
        ("ature", E),
        ("early", Y),
        ("ehood", A),
        ("eless", A),
        ("elily", A),
        ("ement", A),
        ("enced", A),
        ("ences", A),
        ("eness", E),
        ("ening", E),
        ("ental", A),
        ("ented", C),
        ("ently", A),
        ("fully", A),
        ("ially", A),
        ("icant", A),
        ("ician", A),
        ("icide", A),
        ("icism", A),
        ("icist", A),
        ("icity", A),
        ("idine", I),
        ("iedly", A),
        ("ihood", A),
        ("inate", A),
        ("iness", A),
        ("ingly", B),
        ("inism", J),
        ("inity", c),
        ("ional", A),
        ("ioned", A),
        ("ished", A),
        ("istic", A),
        ("ities", A),
        ("itous", A),
        ("ively", A),
        ("ivity", A),
        ("izers", F),
        ("izing", F),
        ("oidal", A),
        ("oides", A),
        ("otide", A),
        ("ously", A)))
m[4] = dict((
        ("able", A),
        ("ably", A),
        ("ages", B),
        ("ally", B),
        ("ance", B),
        ("ancy", B),
        ("ants", B),
        ("aric", A),
        ("arly", K),
        ("ated", I),
        ("ates", A),
        ("atic", B),
        ("ator", A),
        ("ealy", Y),
        ("edly", E),
        ("eful", A),
        ("eity", A),
        ("ence", A),
        ("ency", A),
        ("ened", E),
        ("enly", E),
        ("eous", A),
        ("hood", A),
        ("ials", A),
        ("ians", A),
        ("ible", A),
        ("ibly", A),
        ("ical", A),
        ("ides", L),
        ("iers", A),
        ("iful", A),
        ("ines", M),
        ("ings", N),
        ("ions", B),
        ("ious", A),
        ("isms", B),
        ("ists", A),
        ("itic", H),
        ("ized", F),
        ("izer", F),
        ("less", A),
        ("lily", A),
        ("ness", A),
        ("ogen", A),
        ("ward", A),
        ("wise", A),
        ("ying", B),
        ("yish", A)))
m[3] = dict((
        ("acy", A),
        ("age", B),
        ("aic", A),
        ("als", b),
        ("ant", B),
        ("ars", O),
        ("ary", F),
        ("ata", A),
        ("ate", A),
        ("eal", Y),
        ("ear", Y),
        ("ely", E),
        ("ene", E),
        ("ent", C),
        ("ery", E),
        ("ese", A),
        ("ful", A),
        ("ial", A),
        ("ian", A),
        ("ics", A),
        ("ide", L),
        ("ied", A),
        ("ier", A),
        ("ies", P),
        ("ily", A),
        ("ine", M),
        ("ing", N),
        ("ion", Q),
        ("ish", C),
        ("ism", B),
        ("ist", A),
        ("ite", a),
        ("ity", A),
        ("ium", A),
        ("ive", A),
        ("ize", F),
        ("oid", A),
        ("one", R),
        ("ous", A)))
m[2] = dict((
        ("ae", A),
        ("al", b),
        ("ar", X),
        ("as", B),
        ("ed", E),
        ("en", F),
        ("es", E),
        ("ia", A),
        ("ic", A),
        ("is", A),
        ("ly", B),
        ("on", S),
        ("or", T),
        ("um", U),
        ("us", V),
        ("yl", R),
        ("s'", A),
        ("'s", A)))
m[1] = dict((
        ("a", A),
        ("e", A),
        ("i", A),
        ("o", A),
        ("s", W),
        ("y", B)))


def remove_ending(word):
    length = len(word)
    el = 11
    while el > 0:
        if length - el > 1:
            ending = word[length - el:]
            cond = m[el].get(ending)
            if cond:
                base = word[:length - el]
                if cond(base):
                    return base
        el -= 1
    return word


_endings = (("iev", "ief"),
            ("uct", "uc"),
            ("iev", "ief"),
            ("uct", "uc"),
            ("umpt", "um"),
            ("rpt", "rb"),
            ("urs", "ur"),
            ("istr", "ister"),
            ("metr", "meter"),
            ("olv", "olut"),
            ("ul", "l", "aoi"),
            ("bex", "bic"),
            ("dex", "dic"),
            ("pex", "pic"),
            ("tex", "tic"),
            ("ax", "ac"),
            ("ex", "ec"),
            ("ix", "ic"),
            ("lux", "luc"),
            ("uad", "uas"),
            ("vad", "vas"),
            ("cid", "cis"),
            ("lid", "lis"),
            ("erid", "eris"),
            ("pand", "pans"),
            ("end", "ens", "s"),
            ("ond", "ons"),
            ("lud", "lus"),
            ("rud", "rus"),
            ("her", "hes", "pt"),
            ("mit", "mis"),
            ("ent", "ens", "m"),
            ("ert", "ers"),
            ("et", "es", "n"),
            ("yt", "ys"),
            ("yz", "ys"))


# Hash the ending rules by the last letter of the target ending
_endingrules = defaultdict(list)
for rule in _endings:
    _endingrules[rule[0][-1]].append(rule)

_doubles = frozenset(("dd", "gg", "ll", "mm", "nn", "pp", "rr", "ss", "tt"))


def fix_ending(word):
    if word[-2:] in _doubles:
        word = word[:-1]
    
    for endingrule in _endingrules[word[-1]]:
        target, newend = endingrule[:2]
        if word.endswith(target):
            if len(endingrule) > 2:
                exceptafter = endingrule[2]
                c = word[0 - (len(target) + 1)]
                if c in exceptafter:
                    return word
            
            return word[:0 - len(target)] + newend
    
    return word


def stem(word):
    """Returns the stemmed version of the argument string.
    """
    return fix_ending(remove_ending(word))






#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import os
import tempfile
from multiprocessing import Process, Queue, cpu_count
from cPickle import dump, load

from whoosh.filedb.filetables import LengthWriter, LengthReader
from whoosh.filedb.fileindex import Segment
from whoosh.filedb.filewriting import SegmentWriter
from whoosh.filedb.pools import (imerge, read_run, PoolBase, TempfilePool)
from whoosh.filedb.structfile import StructFile
from whoosh.writing import IndexWriter


# Multiprocessing writer

class SegmentWritingTask(Process):
    def __init__(self, storage, indexname, segname, kwargs, jobqueue,
                 firstjob=None):
        Process.__init__(self)
        self.storage = storage
        self.indexname = indexname
        self.segname = segname
        self.kwargs = kwargs
        self.jobqueue = jobqueue
        self.firstjob = firstjob
        
        self.segment = None
        self.running = True
    
    def _add_file(self, args):
        writer = self.writer
        filename, length = args
        f = open(filename, "rb")
        for _ in xrange(length):
            writer.add_document(**load(f))
        f.close()
        os.remove(filename)
    
    def run(self):
        jobqueue = self.jobqueue
        ix = self.storage.open_index(self.indexname)
        writer = self.writer = SegmentWriter(ix, lock=False, name=self.segname,
                                             **self.kwargs)
        
        if self.firstjob:
            self._add_file(self.firstjob)
        
        while self.running:
            args = jobqueue.get()
            if args is None:
                break
            self._add_file(args)
            
        if not self.running:
            writer.cancel()
        else:
            writer.pool.finish(writer.termswriter, writer.docnum,
                               writer.lengthfile)
            writer.termswriter.close()
            self.jobqueue.put(writer._getsegment())
    
    def cancel(self):
        self.running = False
        

class MultiSegmentWriter(IndexWriter):
    def __init__(self, ix, procs=None, batchsize=100, dir=None, **kwargs):
        self.index = ix
        self.procs = procs or cpu_count()
        self.bufferlimit = batchsize
        self.dir = dir
        self.kwargs = kwargs
        self.kwargs["dir"] = dir
        
        self.segnames = []
        self.tasks = []
        self.jobqueue = Queue()
        self.docbuffer = []
        
        self.writelock = ix.lock("WRITELOCK")
        self.writelock.acquire()
        
        info = ix._read_toc()
        self.schema = info.schema
        self.segment_number = info.segment_counter
        self.generation = info.generation + 1
        self.segments = info.segments
        self.storage = ix.storage
        
    def _new_task(self, firstjob):
        ix = self.index
        self.segment_number += 1
        segmentname = Segment.basename(ix.indexname, self.segment_number)
        task = SegmentWritingTask(ix.storage, ix.indexname, segmentname,
                                  self.kwargs, self.jobqueue, firstjob)
        self.tasks.append(task)
        task.start()
        return task
    
    def _enqueue(self):
        doclist = self.docbuffer
        fd, filename = tempfile.mkstemp(".doclist", dir=self.dir)
        f = os.fdopen(fd, "wb")
        for doc in doclist:
            dump(doc, f, -1)
        f.close()
        args = (filename, len(doclist))
        
        if len(self.tasks) < self.procs:
            self._new_task(args)
        else:
            self.jobqueue.put(args)
        
        self.docbuffer = []
    
    def cancel(self):
        try:
            for task in self.tasks:
                task.cancel()
        finally:
            self.lock.release()
    
    def add_document(self, **fields):
        self.docbuffer.append(fields)
        if len(self.docbuffer) >= self.bufferlimit:
            self._enqueue()
    
    def commit(self, **kwargs):
        try:
            for task in self.tasks:
                self.jobqueue.put(None)
            
            for task in self.tasks:
                task.join()
            
            for task in self.tasks:
                taskseg = self.jobqueue.get()
                self.segments.append(taskseg)
            
            self.jobqueue.close()
            
            from whoosh.filedb.fileindex import _write_toc, _clean_files
            _write_toc(self.storage, self.schema, self.index.indexname,
                       self.generation, self.segment_number, self.segments)
            
            readlock = self.index.lock("READLOCK")
            readlock.acquire(True)
            try:
                _clean_files(self.storage, self.index.indexname,
                             self.generation, self.segments)
            finally:
                readlock.release()
        finally:
            self.writelock.release()


# Multiprocessing pool

class PoolWritingTask(Process):
    def __init__(self, schema, dir, jobqueue, resultqueue, limitmb,
                 firstjob=None):
        Process.__init__(self)
        self.schema = schema
        self.dir = dir
        self.jobqueue = jobqueue
        self.resultqueue = resultqueue
        self.limitmb = limitmb
        self.firstjob = firstjob
    
    def _add_file(self, filename, length):
        subpool = self.subpool
        f = open(filename, "rb")
        for _ in xrange(length):
            code, args = load(f)
            if code == 0:
                subpool.add_content(*args)
            elif code == 1:
                subpool.add_posting(*args)
            elif code == 2:
                subpool.add_field_length(*args)
        f.close()
        os.remove(filename)
    
    def run(self):
        jobqueue = self.jobqueue
        rqueue = self.resultqueue
        subpool = self.subpool = TempfilePool(self.schema, limitmb=self.limitmb,
                                              dir=self.dir)
        
        if self.firstjob:
            self._add_file(*self.firstjob)
        
        while True:
            arg1, arg2 = jobqueue.get()
            if arg1 is None:
                doccount = arg2
                break
            else:
                self._add_file(arg1, arg2)
        
        lenfd, lenfilename = tempfile.mkstemp(".lengths", dir=subpool.dir)
        lenf = os.fdopen(lenfd, "wb")
        subpool._write_lengths(StructFile(lenf), doccount)
        subpool.dump_run()
        rqueue.put((subpool.runs, subpool.fieldlength_totals(),
                    subpool.fieldlength_maxes(), lenfilename))


class MultiPool(PoolBase):
    def __init__(self, schema, dir=None, procs=2, limitmb=32, batchsize=100,
                 **kw):
        PoolBase.__init__(self, schema, dir=dir)
        self._make_dir()
        
        self.procs = procs
        self.limitmb = limitmb
        self.jobqueue = Queue(self.procs * 4)
        self.resultqueue = Queue()
        self.tasks = []
        self.buffer = []
        self.bufferlimit = batchsize
    
    def _new_task(self, firstjob):
        task = PoolWritingTask(self.schema, self.dir, self.jobqueue,
                               self.resultqueue, self.limitmb, firstjob=firstjob)
        self.tasks.append(task)
        task.start()
        return task
    
    def _enqueue(self):
        commandlist = self.buffer
        fd, filename = tempfile.mkstemp(".commands", dir=self.dir)
        f = os.fdopen(fd, "wb")
        for command in commandlist:
            dump(command, f, -1)
        f.close()
        args = (filename, len(commandlist))
        
        if len(self.tasks) < self.procs:
            self._new_task(args)
        else:
            self.jobqueue.put(args)
            
        self.buffer = []
    
    def _append(self, item):
        self.buffer.append(item)
        if len(self.buffer) > self.bufferlimit:
            self._enqueue()
    
    def add_content(self, *args):
        self._append((0, args))
        
    def add_posting(self, *args):
        self.postingqueue.put((1, args))
    
    def add_field_length(self, *args):
        self.postingqueue.put((2, args))
    
    def cancel(self):
        for task in self.tasks:
            task.terminate()
        self.cleanup()
    
    def cleanup(self):
        self._clean_temp_dir()
    
    def finish(self, termswriter, doccount, lengthfile):
        _fieldlength_totals = self._fieldlength_totals
        if not self.tasks:
            return
        
        jobqueue = self.jobqueue
        rqueue = self.resultqueue
        
        for task in self.tasks:
            jobqueue.put((None, doccount))
        
        for task in self.tasks:
            task.join()
        
        runs = []
        lenfilenames = []
        for task in self.tasks:
            taskruns, flentotals, flenmaxes, lenfilename = rqueue.get()
            runs.extend(taskruns)
            lenfilenames.append(lenfilename)
            for fieldnum, total in flentotals.iteritems():
                _fieldlength_totals[fieldnum] += total
            for fieldnum, length in flenmaxes.iteritems():
                if length > self._fieldlength_maxes.get(fieldnum, 0):
                    self._fieldlength_maxes[fieldnum] = length
        
        jobqueue.close()
        rqueue.close()
        
        lw = LengthWriter(lengthfile, doccount)
        for lenfilename in lenfilenames:
            sublengths = LengthReader(StructFile(open(lenfilename, "rb")), doccount)
            lw.add_all(sublengths)
            os.remove(lenfilename)
        lw.close()
        lengths = lw.reader()
        
#        if len(runs) >= self.procs * 2:
#            pool = Pool(self.procs)
#            tempname = lambda: tempfile.mktemp(suffix=".run", dir=self.dir)
#            while len(runs) >= self.procs * 2:
#                runs2 = [(runs[i:i+4], tempname())
#                         for i in xrange(0, len(runs), 4)]
#                if len(runs) % 4:
#                    last = runs2.pop()[0]
#                    runs2[-1][0].extend(last)
#                runs = pool.map(merge_runs, runs2)
#            pool.close()
        
        iterator = imerge([read_run(runname, count) for runname, count in runs])
        total = sum(count for runname, count in runs)
        termswriter.add_iter(iterator, lengths.get)
        for runname, count in runs:
            os.remove(runname)
        
        self.cleanup()
 
 
 
 
 


#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from whoosh.formats import Format
from whoosh.writing import PostingWriter
from whoosh.matching import Matcher, ReadTooFar
from whoosh.spans import Span
from whoosh.system import _INT_SIZE
from whoosh.filedb import postblocks


class FilePostingWriter(PostingWriter):
    blockclass = postblocks.current
    
    def __init__(self, postfile, stringids=False, blocklimit=128,
                 compression=3):
        self.postfile = postfile
        self.stringids = stringids

        if blocklimit > 255:
            raise ValueError("blocklimit argument must be <= 255")
        elif blocklimit < 1:
            raise ValueError("blocklimit argument must be > 0")
        self.blocklimit = blocklimit
        self.compression = compression
        self.block = None

    def _reset_block(self):
        self.block = self.blockclass(self.postfile, self.stringids)
        
    def start(self, format):
        if self.block is not None:
            raise Exception("Called start() in a block")

        self.format = format
        self.blockcount = 0
        self.posttotal = 0
        self.startoffset = self.postfile.tell()
        
        # Magic number
        self.postfile.write_int(self.blockclass.magic)
        # Placeholder for block count
        self.postfile.write_uint(0)
        
        self._reset_block()
        return self.startoffset

    def write(self, id, weight, valuestring, dfl):
        self.block.append(id, weight, valuestring, dfl)
        if len(self.block) >= self.blocklimit:
            self._write_block()
        self.posttotal += 1

    def finish(self):
        if self.block is None:
            raise Exception("Called finish() when not in a block")

        if self.block:
            self._write_block()

        # Seek back to the start of this list of posting blocks and writer the
        # number of blocks
        pf = self.postfile
        pf.flush()
        offset = pf.tell()
        pf.seek(self.startoffset + _INT_SIZE)
        pf.write_uint(self.blockcount)
        pf.seek(offset)
        
        self.block = None
        return self.posttotal

    def cancel(self):
        self.block = None

    def close(self):
        if self.block:
            self.finish()
        self.postfile.close()

    def block_stats(self):
        return self.block.stats()

    def _write_block(self):
        self.block.to_file(self.postfile, self.format.posting_size,
                           compression=self.compression)
        self._reset_block()
        self.blockcount += 1
        
    def as_inline(self):
        block = self.block
        _, maxwol, minlength = block.stats()
        return (tuple(block.ids), tuple(block.weights), tuple(block.values),
                maxwol, minlength)


class FilePostingReader(Matcher):
    def __init__(self, postfile, offset, format, scorer=None,
                 fieldname=None, text=None, stringids=False):
        
        assert isinstance(offset, (int, long)), "offset is %r/%s" % (offset, type(offset))
        assert isinstance(format, Format), "format is %r/%s" % (format, type(format))
        
        self.postfile = postfile
        self.startoffset = offset
        self.format = format
        self.supports_chars = self.format.supports("characters")
        self.supports_poses = self.format.supports("positions")
        self.scorer = scorer
        self.fieldname = fieldname
        self.text = text
        self.stringids = stringids
        
        magic = postfile.get_int(offset)
        self.blockclass = postblocks.magic_map[magic]
        
        self.blockcount = postfile.get_uint(offset + _INT_SIZE)
        self.baseoffset = offset + _INT_SIZE * 2
        self._active = True
        self.currentblock = -1
        self._next_block()

    def __repr__(self):
        return "%s(%r, %s, %r, %r)" % (self.__class__.__name__, str(self.postfile),
                                       self.startoffset, self.fieldname, self.text)

    def close(self):
        pass

    def copy(self):
        return self.__class__(self.postfile, self.startoffset, self.format,
                              scorer=self.scorer, fieldname=self.fieldname,
                              text=self.text, stringids=self.stringids)

    def is_active(self):
        return self._active

    def id(self):
        return self.block.ids[self.i]

    def items_as(self, astype):
        decoder = self.format.decoder(astype)
        for id, value in self.all_items():
            yield (id, decoder(value))

    def supports(self, astype):
        return self.format.supports(astype)

    def value(self):
        if self.block.values is None:
            self.block.read_values(self.format.posting_size)
        return self.block.values[self.i]

    def value_as(self, astype):
        decoder = self.format.decoder(astype)
        return decoder(self.value())

    def spans(self):
        if self.supports_chars:
            return [Span(pos, startchar=startchar, endchar=endchar)
                    for pos, startchar, endchar in self.value_as("characters")]
        elif self.supports_poses:
            return [Span(pos) for pos in self.value_as("positions")]
        else:
            raise Exception("Field does not support positions (%r)" % self.fieldname)

    def weight(self):
        weights = self.block.weights
        if weights is None:
            return 1.0
        else:
            return weights[self.i]
    
    def all_ids(self):
        nextoffset = self.baseoffset
        for _ in xrange(self.blockcount):
            block = self._read_block(nextoffset)
            nextoffset = block.nextoffset
            ids = block.read_ids()
            for id in ids:
                yield id

    def next(self):
        if self.i == self.block.postcount - 1:
            self._next_block()
            return True
        else:
            self.i += 1
            return False

    def skip_to(self, id):
        if not self.is_active():
            raise ReadTooFar
        
        i = self.i
        # If we're already in the block with the target ID, do nothing
        if id <= self.block.ids[i]:
            return
        
        # Skip to the block that would contain the target ID
        if id > self.block.maxid:
            self._skip_to_block(lambda: id > self.block.maxid)
        if not self._active:
            return

        # Iterate through the IDs in the block until we find or pass the
        # target
        ids = self.block.ids
        i = self.i
        while ids[i] < id:
            i += 1
            if i == len(ids):
                self._active = False
                return
        self.i = i

    def _read_block(self, offset):
        pf = self.postfile
        pf.seek(offset)
        return self.blockclass.from_file(pf, self.stringids)
        
    def _consume_block(self):
        self.block.read_ids()
        self.block.read_weights()
        self.i = 0

    def _next_block(self, consume=True):
        if not (self.currentblock < self.blockcount):
            raise Exception("No next block")
        
        self.currentblock += 1
        if self.currentblock == self.blockcount:
            self._active = False
            return

        if self.currentblock == 0:
            pos = self.baseoffset
        else:
            pos = self.block.nextoffset

        self.block = self._read_block(pos)
        if consume:
            self._consume_block()

    def _skip_to_block(self, targetfn):
        skipped = 0
        while self._active and targetfn():
            self._next_block(consume=False)
            skipped += 1

        if self._active:
            self._consume_block()
        
        return skipped
    
    def supports_quality(self):
        return self.scorer and self.scorer.supports_quality()
    
    def skip_to_quality(self, minquality):
        bq = self.block_quality
        if bq() > minquality:
            return 0
        return self._skip_to_block(lambda: bq() <= minquality)
    
    def block_maxweight(self):
        return self.block.maxweight
    
    def block_maxwol(self):
        return self.block.maxwol
    
    def block_maxid(self):
        return self.block.maxid
    
    def block_minlength(self):
        return self.block.minlength
    
    def score(self):
        return self.scorer.score(self)
    
    def quality(self):
        return self.scorer.quality(self)
    
    def block_quality(self):
        return self.scorer.block_quality(self)
    
    
    
    
        












